# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class StreamGenerationEnd(UniversalBaseModel):
    """
    The end of generation. There may still be more information such as the
    factual consistency score, but generation has stopped.
    """

    type: typing.Literal["generation_end"] = "generation_end"

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from .data import Data
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class Table(UniversalBaseModel):
    """
    A table in a document.
    """

    id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The unique ID of the table within the document.
    """

    title: typing.Optional[str] = pydantic.Field(default=None)
    """
    The title of the table.
    """

    data: typing.Optional[Data] = pydantic.Field(default=None)
    """
    The data of the table.
    """

    description: typing.Optional[str] = pydantic.Field(default=None)
    """
    The description of the table.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class ChatInfoResponse(UniversalBaseModel):
    """
    Information about the chat.
    """

    type: typing.Literal["chat_info"] = "chat_info"
    chat_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    ID of the chat.
    """

    turn_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    ID of the turn.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .header import Header
import pydantic
from .row import Row
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class Data(UniversalBaseModel):
    """
    The data of a table.
    """

    headers: typing.Optional[typing.List[Header]] = pydantic.Field(default=None)
    """
    The headers of the table.
    """

    rows: typing.Optional[typing.List[Row]] = pydantic.Field(default=None)
    """
    The rows in the data.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .citation_parameters_style import CitationParametersStyle
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class CitationParameters(UniversalBaseModel):
    """
    Style the generator should use when making citations.
    """

    style: typing.Optional[CitationParametersStyle] = pydantic.Field(default=None)
    """
    The citation style to be used in summary.
    Can be one of:
    * `numeric` - Citations formatted as simple numerals: \[1\], \[2\] ...
    * `none` - Citations removed from text.
    * `html` - Citation formatted as a URL like `<a href="url_pattern">text_pattern</a>`.
    * `markdown` - Formatted as `[text_pattern](url_pattern)`.
    """

    url_pattern: typing.Optional[str] = pydantic.Field(default=None)
    """
    The URL pattern if the citation_style is set to `html` or `markdown`.
    The pattern can access metadata attributes in the document or part.
    e.g. `https://my.doc/foo/{doc.id}/{part.id}`
    
    The default `url_pattern` is an empty string.
    """

    text_pattern: typing.Optional[str] = pydantic.Field(default=None)
    """
    The text pattern if the citation_style is set to `html` or `markdown`.
    This pattern sets the href for HTML or the text within `[]` in markdown,
    and defaults to N being the index of result if it is not set.
    
    The default citation style looks like `[N](<url_pattern>)` for markdown.
    
    You can use metadata attributes in the `text_pattern`. For example,
    the pattern `{doc.title}` with citation style `markdown` would result
    in final citation output like `[Title](<url_pattern>)` when
    the document's metadata includes `{"title":"Title"}`.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class Cell(UniversalBaseModel):
    """
    A cell in a table.
    """

    text_value: typing.Optional[str] = pydantic.Field(default=None)
    """
    A text value.
    """

    int_value: typing.Optional[int] = pydantic.Field(default=None)
    """
    A signed 64-bit integer value.
    """

    float_value: typing.Optional[float] = pydantic.Field(default=None)
    """
    A floating-point value with double precision.
    """

    bool_value: typing.Optional[bool] = pydantic.Field(default=None)
    """
    A boolean value.
    """

    colspan: typing.Optional[int] = pydantic.Field(default=None)
    """
    (Optional) The number of columns the cell spans. Default is 1. Must be greater than 0.
    """

    rowspan: typing.Optional[int] = pydantic.Field(default=None)
    """
    (Optional) The number of rows the cell spans. Default is 1. Must be greater than 0.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing

FilterAttributeType = typing.Union[
    typing.Literal["integer", "real_number", "text", "boolean", "list[integer]", "list[real_number]", "list[text]"],
    typing.Any,
]



# This file was auto-generated by Fern from our API Definition.

import typing

JobState = typing.Union[
    typing.Literal["unknown", "queued", "started", "completed", "failed", "failed_will_retry", "aborted"], typing.Any
]



# This file was auto-generated by Fern from our API Definition.

from ..core.client_wrapper import SyncClientWrapper
import typing
from ..core.request_options import RequestOptions
from ..core.pagination import SyncPager
from ..types.llm import Llm
from ..types.list_ll_ms_response import ListLlMsResponse
from ..core.pydantic_utilities import parse_obj_as
from ..errors.forbidden_error import ForbiddenError
from ..types.error import Error
from json.decoder import JSONDecodeError
from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper
from ..core.pagination import AsyncPager


class LlmsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def list(
        self,
        *,
        filter: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        page_key: typing.Optional[str] = None,
        request_timeout: typing.Optional[int] = None,
        request_timeout_millis: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> SyncPager[Llm]:
        """
        List LLMs that can be used with query and chat endpoints. The LLM is not directly specified in a query,
        but instead a `generation_preset_name` is used. The `generation_preset_name` property in generation parameters
        can be found as the `name` property on the Generations Presets retrieved from `/v2/generation_presets`.

        Parameters
        ----------
        filter : typing.Optional[str]
            A regular expression to match names and descriptions of the LLMs.

        limit : typing.Optional[int]
            The maximum number of results to return in the list.

        page_key : typing.Optional[str]
            Used to retrieve the next page of LLMs after the limit has been reached.
            This parameter is not needed for the first page of results.

        request_timeout : typing.Optional[int]
            The API will make a best effort to complete the request in the specified seconds or time out.

        request_timeout_millis : typing.Optional[int]
            The API will make a best effort to complete the request in the specified milliseconds or time out.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        SyncPager[Llm]
            List of LLMs.

        Examples
        --------
        from vectara import Vectara

        client = Vectara(
            api_key="YOUR_API_KEY",
            client_id="YOUR_CLIENT_ID",
            client_secret="YOUR_CLIENT_SECRET",
        )
        response = client.llms.list()
        for item in response:
            yield item
        # alternatively, you can paginate page-by-page
        for page in response.iter_pages():
            yield page
        """
        _response = self._client_wrapper.httpx_client.request(
            "v2/llms",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "filter": filter,
                "limit": limit,
                "page_key": page_key,
            },
            headers={
                "Request-Timeout": str(request_timeout) if request_timeout is not None else None,
                "Request-Timeout-Millis": str(request_timeout_millis) if request_timeout_millis is not None else None,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _parsed_response = typing.cast(
                    ListLlMsResponse,
                    parse_obj_as(
                        type_=ListLlMsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                _has_next = False
                _get_next = None
                if _parsed_response.metadata is not None:
                    _parsed_next = _parsed_response.metadata.page_key
                    _has_next = _parsed_next is not None and _parsed_next != ""
                    _get_next = lambda: self.list(
                        filter=filter,
                        limit=limit,
                        page_key=_parsed_next,
                        request_timeout=request_timeout,
                        request_timeout_millis=request_timeout_millis,
                        request_options=request_options,
                    )
                _items = _parsed_response.llms
                return SyncPager(has_next=_has_next, items=_items, get_next=_get_next)
            if _response.status_code == 403:
                raise ForbiddenError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncLlmsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def list(
        self,
        *,
        filter: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        page_key: typing.Optional[str] = None,
        request_timeout: typing.Optional[int] = None,
        request_timeout_millis: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncPager[Llm]:
        """
        List LLMs that can be used with query and chat endpoints. The LLM is not directly specified in a query,
        but instead a `generation_preset_name` is used. The `generation_preset_name` property in generation parameters
        can be found as the `name` property on the Generations Presets retrieved from `/v2/generation_presets`.

        Parameters
        ----------
        filter : typing.Optional[str]
            A regular expression to match names and descriptions of the LLMs.

        limit : typing.Optional[int]
            The maximum number of results to return in the list.

        page_key : typing.Optional[str]
            Used to retrieve the next page of LLMs after the limit has been reached.
            This parameter is not needed for the first page of results.

        request_timeout : typing.Optional[int]
            The API will make a best effort to complete the request in the specified seconds or time out.

        request_timeout_millis : typing.Optional[int]
            The API will make a best effort to complete the request in the specified milliseconds or time out.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncPager[Llm]
            List of LLMs.

        Examples
        --------
        import asyncio

        from vectara import AsyncVectara

        client = AsyncVectara(
            api_key="YOUR_API_KEY",
            client_id="YOUR_CLIENT_ID",
            client_secret="YOUR_CLIENT_SECRET",
        )


        async def main() -> None:
            response = await client.llms.list()
            async for item in response:
                yield item
            # alternatively, you can paginate page-by-page
            async for page in response.iter_pages():
                yield page


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v2/llms",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "filter": filter,
                "limit": limit,
                "page_key": page_key,
            },
            headers={
                "Request-Timeout": str(request_timeout) if request_timeout is not None else None,
                "Request-Timeout-Millis": str(request_timeout_millis) if request_timeout_millis is not None else None,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _parsed_response = typing.cast(
                    ListLlMsResponse,
                    parse_obj_as(
                        type_=ListLlMsResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                _has_next = False
                _get_next = None
                if _parsed_response.metadata is not None:
                    _parsed_next = _parsed_response.metadata.page_key
                    _has_next = _parsed_next is not None and _parsed_next != ""
                    _get_next = lambda: self.list(
                        filter=filter,
                        limit=limit,
                        page_key=_parsed_next,
                        request_timeout=request_timeout,
                        request_timeout_millis=request_timeout_millis,
                        request_options=request_options,
                    )
                _items = _parsed_response.llms
                return AsyncPager(has_next=_has_next, items=_items, get_next=_get_next)
            if _response.status_code == 403:
                raise ForbiddenError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



import json
import logging
from pathlib import Path
from typing import Union, Optional, Dict
from vectara.upload.client import UploadClient
from vectara.types import Document
import mimetypes


class UploadManager:

    def __init__(self, upload_client: UploadClient):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.upload_client = upload_client

    def _discover_mime_type(self, name: str):
        mime_type, encoding = mimetypes.guess_type(name)
        return mime_type

    def upload(self, corpus_key: str, target: Union[str, Path], doc_id: Union[None, str] = None,
               metadata: Optional[Dict] = None) -> Document:

        if isinstance(target, str):
            target = Path(target)

        # TODO Add directory processing.
        if not doc_id:
            doc_id = target.name

        content_type = self._discover_mime_type(target.name)  # Change this based on extension.
        with open(target, "rb") as f:
            content = f.read()

        return self.upload_client.file(corpus_key, file=(doc_id, content, content_type), metadata=metadata)









# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import typing


class TableExtractionConfig(UniversalBaseModel):
    """
    (Optional) Configuration for table extraction from the document.
    """

    extract_tables: bool = pydantic.Field()
    """
    If set to true, the platform will attempt to extract tables from the document.
    The tables will be indexed as separate document parts.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class UserFunctionReranker(UniversalBaseModel):
    type: typing.Literal["userfn"] = "userfn"
    user_function: typing.Optional[str] = pydantic.Field(default=None)
    """
    The user defined function.
    """

    limit: typing.Optional[int] = pydantic.Field(default=None)
    """
    Specifies the maximum number of results to be returned after the reranking process. 
    When a reranker is applied, it performs the following steps:
    1. Reranks all input results according to its algorithm.
    2. Sorts the reranked results based on their new scores.
    3. Returns the top N results, where N is the value specified by this limit.
    
    Note: This limit is applied per reranking stage. In a chain of rerankers, 
    each reranker can have its own limit, potentially reducing the number of 
    results at each stage.
    """

    cutoff: typing.Optional[float] = pydantic.Field(default=None)
    """
    Specifies the minimum score threshold for results to be included after the reranking process.
    When a reranker is applied with a cutoff, it performs the following steps:
    1. Reranks all input results according to its algorithm.
    2. Applies the cutoff, removing any results with scores below the specified threshold.
    3. Returns the remaining results, sorted by their new scores.
    
    Note: This cutoff is applied per reranking stage. In a chain of rerankers,
    each reranker can have its own cutoff, potentially further reducing the number of
    results at each stage. If both 'limit' and 'cutoff' are specified, the cutoff
    is applied first, followed by the limit.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from .api_key import ApiKey
from .api_key_role import ApiKeyRole
from .api_operation_policy import ApiOperationPolicy
from .api_policy import ApiPolicy
from .api_role import ApiRole
from .app_client import AppClient
from .bad_request_error_body import BadRequestErrorBody
from .cell import Cell
from .chain_reranker import ChainReranker
from .chat import Chat
from .chat_full_response import ChatFullResponse
from .chat_info_response import ChatInfoResponse
from .chat_parameters import ChatParameters
from .chat_streamed_response import ChatStreamedResponse
from .citation_parameters import CitationParameters
from .citation_parameters_style import CitationParametersStyle
from .components_schemas_create_client_credentials_request import ComponentsSchemasCreateClientCredentialsRequest
from .components_schemas_max_chars_chunking_strategy import ComponentsSchemasMaxCharsChunkingStrategy
from .context_configuration import ContextConfiguration
from .core_document import CoreDocument
from .core_document_part import CoreDocumentPart
from .corpus import Corpus
from .corpus_custom_dimension import CorpusCustomDimension
from .corpus_key import CorpusKey
from .corpus_limits import CorpusLimits
from .create_client_credentials_request import CreateClientCredentialsRequest
from .create_document_request import CreateDocumentRequest
from .custom_dimensions import CustomDimensions
from .customer_specific_reranker import CustomerSpecificReranker
from .data import Data
from .document import Document
from .document_part import DocumentPart
from .document_storage_usage import DocumentStorageUsage
from .encoder import Encoder
from .error import Error
from .extraction_usage import ExtractionUsage
from .factual_consistency_score import FactualConsistencyScore
from .factual_consistency_score_span import FactualConsistencyScoreSpan
from .filter_attribute import FilterAttribute
from .filter_attribute_level import FilterAttributeLevel
from .filter_attribute_type import FilterAttributeType
from .generation_info import GenerationInfo
from .generation_parameters import GenerationParameters
from .generation_preset import GenerationPreset
from .generation_span import GenerationSpan
from .header import Header
from .individual_search_result import IndividualSearchResult
from .job import Job
from .job_state import JobState
from .job_type import JobType
from .keyed_search_corpus import KeyedSearchCorpus
from .language import Language
from .list_api_keys_response import ListApiKeysResponse
from .list_app_clients_response import ListAppClientsResponse
from .list_chat_turns_response import ListChatTurnsResponse
from .list_chats_response import ListChatsResponse
from .list_corpora_response import ListCorporaResponse
from .list_documents_response import ListDocumentsResponse
from .list_encoders_response import ListEncodersResponse
from .list_generation_presets_response import ListGenerationPresetsResponse
from .list_jobs_response import ListJobsResponse
from .list_ll_ms_response import ListLlMsResponse
from .list_metadata import ListMetadata
from .list_query_histories_response import ListQueryHistoriesResponse
from .list_query_histories_response_metadata import ListQueryHistoriesResponseMetadata
from .list_rerankers_response import ListRerankersResponse
from .list_users_response import ListUsersResponse
from .llm import Llm
from .max_chars_chunking_strategy import MaxCharsChunkingStrategy
from .mmr_reranker import MmrReranker
from .model_parameters import ModelParameters
from .none_reranker import NoneReranker
from .not_found_error_body import NotFoundErrorBody
from .prompt import Prompt
from .query_full_response import QueryFullResponse
from .query_history import QueryHistory
from .query_history_span import QueryHistorySpan
from .query_history_summary import QueryHistorySummary
from .query_streamed_response import QueryStreamedResponse
from .query_warning import QueryWarning
from .rephrase_span import RephraseSpan
from .replace_filter_attributes_response import ReplaceFilterAttributesResponse
from .rerank_span import RerankSpan
from .reranked_search_result import RerankedSearchResult
from .reranker import Reranker
from .row import Row
from .search_corpora_parameters import SearchCorporaParameters
from .search_corpus import SearchCorpus
from .search_parameters import SearchParameters
from .search_reranker import SearchReranker
from .search_semantics import SearchSemantics
from .search_span import SearchSpan
from .stream_error import StreamError
from .stream_generation_chunk import StreamGenerationChunk
from .stream_generation_end import StreamGenerationEnd
from .stream_response_end import StreamResponseEnd
from .stream_search_response import StreamSearchResponse
from .structured_document import StructuredDocument
from .structured_document_section import StructuredDocumentSection
from .table import Table
from .table_extraction_config import TableExtractionConfig
from .turn import Turn
from .update_document_request import UpdateDocumentRequest
from .user import User
from .user_function_reranker import UserFunctionReranker

__all__ = [
    "ApiKey",
    "ApiKeyRole",
    "ApiOperationPolicy",
    "ApiPolicy",
    "ApiRole",
    "AppClient",
    "BadRequestErrorBody",
    "Cell",
    "ChainReranker",
    "Chat",
    "ChatFullResponse",
    "ChatInfoResponse",
    "ChatParameters",
    "ChatStreamedResponse",
    "CitationParameters",
    "CitationParametersStyle",
    "ComponentsSchemasCreateClientCredentialsRequest",
    "ComponentsSchemasMaxCharsChunkingStrategy",
    "ContextConfiguration",
    "CoreDocument",
    "CoreDocumentPart",
    "Corpus",
    "CorpusCustomDimension",
    "CorpusKey",
    "CorpusLimits",
    "CreateClientCredentialsRequest",
    "CreateDocumentRequest",
    "CustomDimensions",
    "CustomerSpecificReranker",
    "Data",
    "Document",
    "DocumentPart",
    "DocumentStorageUsage",
    "Encoder",
    "Error",
    "ExtractionUsage",
    "FactualConsistencyScore",
    "FactualConsistencyScoreSpan",
    "FilterAttribute",
    "FilterAttributeLevel",
    "FilterAttributeType",
    "GenerationInfo",
    "GenerationParameters",
    "GenerationPreset",
    "GenerationSpan",
    "Header",
    "IndividualSearchResult",
    "Job",
    "JobState",
    "JobType",
    "KeyedSearchCorpus",
    "Language",
    "ListApiKeysResponse",
    "ListAppClientsResponse",
    "ListChatTurnsResponse",
    "ListChatsResponse",
    "ListCorporaResponse",
    "ListDocumentsResponse",
    "ListEncodersResponse",
    "ListGenerationPresetsResponse",
    "ListJobsResponse",
    "ListLlMsResponse",
    "ListMetadata",
    "ListQueryHistoriesResponse",
    "ListQueryHistoriesResponseMetadata",
    "ListRerankersResponse",
    "ListUsersResponse",
    "Llm",
    "MaxCharsChunkingStrategy",
    "MmrReranker",
    "ModelParameters",
    "NoneReranker",
    "NotFoundErrorBody",
    "Prompt",
    "QueryFullResponse",
    "QueryHistory",
    "QueryHistorySpan",
    "QueryHistorySummary",
    "QueryStreamedResponse",
    "QueryWarning",
    "RephraseSpan",
    "ReplaceFilterAttributesResponse",
    "RerankSpan",
    "RerankedSearchResult",
    "Reranker",
    "Row",
    "SearchCorporaParameters",
    "SearchCorpus",
    "SearchParameters",
    "SearchReranker",
    "SearchSemantics",
    "SearchSpan",
    "StreamError",
    "StreamGenerationChunk",
    "StreamGenerationEnd",
    "StreamResponseEnd",
    "StreamSearchResponse",
    "StructuredDocument",
    "StructuredDocumentSection",
    "Table",
    "TableExtractionConfig",
    "Turn",
    "UpdateDocumentRequest",
    "User",
    "UserFunctionReranker",
]



# This file was auto-generated by Fern from our API Definition.

import typing
from ..core.client_wrapper import SyncClientWrapper
from ..types.corpus_key import CorpusKey
from ..core.request_options import RequestOptions
from ..types.document import Document
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..errors.forbidden_error import ForbiddenError
from ..types.error import Error
from ..errors.not_found_error import NotFoundError
from ..types.not_found_error_body import NotFoundErrorBody
from ..errors.too_many_requests_error import TooManyRequestsError
from json.decoder import JSONDecodeError
from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class IndexClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def update_corpus_document(
        self,
        corpus_key: CorpusKey,
        document_id: str,
        *,
        request_timeout: typing.Optional[int] = None,
        request_timeout_millis: typing.Optional[int] = None,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Document:
        """
        Updates document identified by its unique `document_id` from a specific
        corpus. The request body metadata is merged with the existing metadata,
        adding or modifying only the specified fields.

        Parameters
        ----------
        corpus_key : CorpusKey
            The unique key identifying the corpus with the document to update.

        document_id : str
            The document ID of the document to update.
            This `document_id` must be percent encoded.

        request_timeout : typing.Optional[int]
            The API will make a best effort to complete the request in the specified seconds or time out.

        request_timeout_millis : typing.Optional[int]
            The API will make a best effort to complete the request in the specified milliseconds or time out.

        metadata : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]
            The metadata for a document as an arbitrary object. Properties of this object
            can be used by document level filter attributes.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Document
            Successfully updated the document.

        Examples
        --------
        from vectara import Vectara

        client = Vectara(
            api_key="YOUR_API_KEY",
            client_id="YOUR_CLIENT_ID",
            client_secret="YOUR_CLIENT_SECRET",
        )
        client.index.update_corpus_document(
            corpus_key="my-corpus",
            document_id="document_id",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v2/corpora/{jsonable_encoder(corpus_key)}/documents/{jsonable_encoder(document_id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="PATCH",
            json={
                "metadata": metadata,
            },
            headers={
                "Request-Timeout": str(request_timeout) if request_timeout is not None else None,
                "Request-Timeout-Millis": str(request_timeout_millis) if request_timeout_millis is not None else None,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    Document,
                    parse_obj_as(
                        type_=Document,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def replace_corpus_document_metadata(
        self,
        corpus_key: CorpusKey,
        document_id: str,
        *,
        request_timeout: typing.Optional[int] = None,
        request_timeout_millis: typing.Optional[int] = None,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Document:
        """
        Replaces metadata of a document identified by its unique `document_id`
        from a specific corpus.

        Parameters
        ----------
        corpus_key : CorpusKey
            The unique key identifying the corpus with the document to update.

        document_id : str
            The document ID of the document to update.
            This `document_id` must be percent encoded.

        request_timeout : typing.Optional[int]
            The API will make a best effort to complete the request in the specified seconds or time out.

        request_timeout_millis : typing.Optional[int]
            The API will make a best effort to complete the request in the specified milliseconds or time out.

        metadata : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]
            The metadata for a document as an arbitrary object. Properties of this object
            can be used by document level filter attributes.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Document
            Successfully updated the document.

        Examples
        --------
        from vectara import Vectara

        client = Vectara(
            api_key="YOUR_API_KEY",
            client_id="YOUR_CLIENT_ID",
            client_secret="YOUR_CLIENT_SECRET",
        )
        client.index.replace_corpus_document_metadata(
            corpus_key="my-corpus",
            document_id="document_id",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v2/corpora/{jsonable_encoder(corpus_key)}/documents/{jsonable_encoder(document_id)}/metadata",
            base_url=self._client_wrapper.get_environment().default,
            method="PUT",
            json={
                "metadata": metadata,
            },
            headers={
                "Request-Timeout": str(request_timeout) if request_timeout is not None else None,
                "Request-Timeout-Millis": str(request_timeout_millis) if request_timeout_millis is not None else None,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    Document,
                    parse_obj_as(
                        type_=Document,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncIndexClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def update_corpus_document(
        self,
        corpus_key: CorpusKey,
        document_id: str,
        *,
        request_timeout: typing.Optional[int] = None,
        request_timeout_millis: typing.Optional[int] = None,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Document:
        """
        Updates document identified by its unique `document_id` from a specific
        corpus. The request body metadata is merged with the existing metadata,
        adding or modifying only the specified fields.

        Parameters
        ----------
        corpus_key : CorpusKey
            The unique key identifying the corpus with the document to update.

        document_id : str
            The document ID of the document to update.
            This `document_id` must be percent encoded.

        request_timeout : typing.Optional[int]
            The API will make a best effort to complete the request in the specified seconds or time out.

        request_timeout_millis : typing.Optional[int]
            The API will make a best effort to complete the request in the specified milliseconds or time out.

        metadata : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]
            The metadata for a document as an arbitrary object. Properties of this object
            can be used by document level filter attributes.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Document
            Successfully updated the document.

        Examples
        --------
        import asyncio

        from vectara import AsyncVectara

        client = AsyncVectara(
            api_key="YOUR_API_KEY",
            client_id="YOUR_CLIENT_ID",
            client_secret="YOUR_CLIENT_SECRET",
        )


        async def main() -> None:
            await client.index.update_corpus_document(
                corpus_key="my-corpus",
                document_id="document_id",
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v2/corpora/{jsonable_encoder(corpus_key)}/documents/{jsonable_encoder(document_id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="PATCH",
            json={
                "metadata": metadata,
            },
            headers={
                "Request-Timeout": str(request_timeout) if request_timeout is not None else None,
                "Request-Timeout-Millis": str(request_timeout_millis) if request_timeout_millis is not None else None,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    Document,
                    parse_obj_as(
                        type_=Document,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def replace_corpus_document_metadata(
        self,
        corpus_key: CorpusKey,
        document_id: str,
        *,
        request_timeout: typing.Optional[int] = None,
        request_timeout_millis: typing.Optional[int] = None,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Document:
        """
        Replaces metadata of a document identified by its unique `document_id`
        from a specific corpus.

        Parameters
        ----------
        corpus_key : CorpusKey
            The unique key identifying the corpus with the document to update.

        document_id : str
            The document ID of the document to update.
            This `document_id` must be percent encoded.

        request_timeout : typing.Optional[int]
            The API will make a best effort to complete the request in the specified seconds or time out.

        request_timeout_millis : typing.Optional[int]
            The API will make a best effort to complete the request in the specified milliseconds or time out.

        metadata : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]
            The metadata for a document as an arbitrary object. Properties of this object
            can be used by document level filter attributes.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Document
            Successfully updated the document.

        Examples
        --------
        import asyncio

        from vectara import AsyncVectara

        client = AsyncVectara(
            api_key="YOUR_API_KEY",
            client_id="YOUR_CLIENT_ID",
            client_secret="YOUR_CLIENT_SECRET",
        )


        async def main() -> None:
            await client.index.replace_corpus_document_metadata(
                corpus_key="my-corpus",
                document_id="document_id",
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v2/corpora/{jsonable_encoder(corpus_key)}/documents/{jsonable_encoder(document_id)}/metadata",
            base_url=self._client_wrapper.get_environment().default,
            method="PUT",
            json={
                "metadata": metadata,
            },
            headers={
                "Request-Timeout": str(request_timeout) if request_timeout is not None else None,
                "Request-Timeout-Millis": str(request_timeout_millis) if request_timeout_millis is not None else None,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    Document,
                    parse_obj_as(
                        type_=Document,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 429:
                raise TooManyRequestsError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
import datetime as dt
from .query_history_span import QueryHistorySpan
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class QueryHistory(UniversalBaseModel):
    id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The ID of the query history.
    """

    query: typing.Optional[typing.Optional[typing.Any]] = None
    chat_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The ID of the chat the query is a part of.
    """

    latency_millis: typing.Optional[int] = pydantic.Field(default=None)
    """
    Time taken to complete the query, measured in milliseconds.
    """

    started_at: typing.Optional[dt.datetime] = pydantic.Field(default=None)
    """
    ISO date time indicating when the query was first received.
    """

    spans: typing.Optional[typing.List[QueryHistorySpan]] = pydantic.Field(default=None)
    """
    Parts of the query pipeline. Each span explains what happened during that stage of the query pipeline.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from .max_chars_chunking_strategy import MaxCharsChunkingStrategy

ComponentsSchemasMaxCharsChunkingStrategy = MaxCharsChunkingStrategy



# This file was auto-generated by Fern from our API Definition.

import typing
from .rephrase_span import RephraseSpan
from .search_span import SearchSpan
from .rerank_span import RerankSpan
from .generation_span import GenerationSpan
from .factual_consistency_score_span import FactualConsistencyScoreSpan

QueryHistorySpan = typing.Union[RephraseSpan, SearchSpan, RerankSpan, GenerationSpan, FactualConsistencyScoreSpan]



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class UpdateDocumentRequest(UniversalBaseModel):
    """
    Schema for updating the document. For PUT requests, the request body metadata replaces the existing
    metadata. For PATCH requests, the request body metadata is merged with the existing metadata,
    adding or modifying only the specified fields.
    """

    metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = pydantic.Field(default=None)
    """
    The metadata for a document as an arbitrary object. Properties of this object
    can be used by document level filter attributes.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class CorpusLimits(UniversalBaseModel):
    used_docs: typing.Optional[int] = pydantic.Field(default=None)
    """
    The number of documents contained in the corpus.
    """

    used_parts: typing.Optional[int] = pydantic.Field(default=None)
    """
    The number of document parts contained in the corpus.
    """

    used_bytes: typing.Optional[int] = pydantic.Field(default=None)
    """
    NOTE: This field is currently not populated by the system.
    The number of bytes contained in the corpus. This includes the document metadata,
    document part metadata, and document contents.
    """

    used_characters: typing.Optional[int] = pydantic.Field(default=None)
    """
    The number of characters contained in the corpus. This includes the document metadata,
    document part metadata, and document contents.
    """

    max_bytes: typing.Optional[int] = pydantic.Field(default=None)
    """
    NOTE: This field is currently not populated by the system.
    The maximum number of bytes the corpus can be.
    """

    max_metadata_bytes: typing.Optional[int] = pydantic.Field(default=None)
    """
    The maximum size that metadata can be on documents.
    """

    index_rate: typing.Optional[int] = pydantic.Field(default=None)
    """
    NOTE: This field is currently not populated by the system.
    The maximum per-second addition of new documents to corpus.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.api_error import ApiError
from ..types.not_found_error_body import NotFoundErrorBody


class NotFoundError(ApiError):
    def __init__(self, body: NotFoundErrorBody):
        super().__init__(status_code=404, body=body)



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
import datetime as dt
from .api_role import ApiRole
from .api_policy import ApiPolicy
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class User(UniversalBaseModel):
    id: typing.Optional[str] = pydantic.Field(default=None)
    """
    The Vectara ID for the user.
    """

    email: typing.Optional[str] = pydantic.Field(default=None)
    """
    The email address for the user.
    """

    username: typing.Optional[str] = pydantic.Field(default=None)
    """
    The username for the user.
    """

    enabled: typing.Optional[bool] = pydantic.Field(default=None)
    """
    Indicates whether the user is enabled or disabled.
    """

    description: typing.Optional[str] = pydantic.Field(default=None)
    """
    The description of the user.
    """

    created_at: typing.Optional[dt.datetime] = pydantic.Field(default=None)
    """
    When the user was created.
    """

    updated_at: typing.Optional[dt.datetime] = pydantic.Field(default=None)
    """
    When a user property was last updated.
    """

    api_roles: typing.Optional[typing.List[ApiRole]] = pydantic.Field(default=None)
    """
    The role names of the user.
    """

    api_policy: typing.Optional[ApiPolicy] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from vectara.managers import CreateCorpusRequest, CorpusManager
from vectara.corpora.client import CorporaClient
from vectara.config import ClientConfig, HomeConfigLoader, ApiKeyAuthConfig, OAuth2AuthConfig
from vectara.types import Corpus, QueryFullResponse
from typing import Union, Optional, List, Dict
from abc import ABC
import logging
import os
import re
import getpass
import json


class LabHelper:

    MAX_CORPUS_KEY: int = 50
    MAX_USERNAME_LENGTH: int = 20

    def __init__(self, corpus_manager: CorpusManager):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.corpus_manager = corpus_manager

    # Define a function to convert a string to camel case
    def _camel_case(self, s: str):
        # Use regular expression substitution to replace underscores and hyphens with spaces,
        # then title case the string (capitalize the first letter of each word), and remove spaces
        s = re.sub(r"(_|-)+", " ", s).title().replace(" ", "")

        # Join the string, ensuring the first letter is lowercase
        return ''.join([s[0].lower(), s[1:]])

    def _discover_user(self) -> str:
        username = getpass.getuser()
        if not username:
            raise Exception("Could not extract username from environment prefix")
        else:
            return username

    def _build_lab_name_and_key(self, name: Optional[str], key: Optional[str] = None, user_prefix: bool = True,
                                username=None):
        if not name:
            raise Exception("You must specify a corpus name for the lab")

        if not key:
            key = re.sub('[^0-9a-zA-Z]+', '_', name).lower()

        # First create the lab name which we will use as a filter.
        if user_prefix:
            if not username:
                username = self._discover_user()
            self.logger.info(f"Found username from environment [{username}]")
            # Replace domain
            username = re.sub('@.*', '', username)

            # Replace non-alpha numeric letters with an underscore

            username = re.sub('[^0-9a-zA-Z]+', '_', username)
            if username.endswith("_"):
                username = username[0:(self.MAX_USERNAME_LENGTH - 1)]

            username = self._camel_case(username)[0:self.MAX_USERNAME_LENGTH]

            self.logger.info(f"Converted username is [{username}]")
            name = (username + " - " + name)
            # corpus.key will be validated and set by this point so we can ignore mypy
            key = username + "_" + key[0:self.MAX_CORPUS_KEY] # type: ignore

        self.logger.info(f"Lab corpus name will be [{name}]")
        self.logger.info(f"Lab corpus key will be [{key}]")
        return name, key


    def create_lab_corpus(self, corpus: CreateCorpusRequest, user_prefix=True, username=None) -> Corpus:
        """
        This will create a unique lab based based on the corpus in the request, however it will
        ensure the lab has a unique name and corpus key. This is important as the create/delete cycle
        with a re-used key can cause a race condition - so instead we make sure to create a new unique corpus
        which will have a unique name and corpus_key.

        the corpus in the request should have:
        * A name which is not longer than 50 characters.
        * A corpus key of the format lab_(id), where id is < 10 characters and all lower case / underscore

        This method will modify those properties, adding a user_prefix (if True) and adding a unique postfix
        based on existing labs with the same name.

        :param corpus:
        :param user_prefix:
        :return:
        """

        corpus_clone = corpus.copy()

        name, key = self._build_lab_name_and_key(corpus.name, corpus.key)
        corpus_clone.name = name
        corpus_clone.key = key


        self.logger.info("Creating lab corpus")
        response = self.corpus_manager.create_corpus(corpus_clone, delete_existing=True) # type: ignore
        return response

    @classmethod
    def valid_bool(cls, user_input: str):
        if user_input == 'y' or user_input == 'n':
            return True
        else:
            print("Invalid value, must be [y] or [n].")
            return False

    @classmethod
    def valid_int(cls, value: str, min_value=None, max_value=None) -> bool:
        logger = logging.getLogger(cls.__class__.__name__)

        value_int: int = -1
        try:
            value_int = int(value)
        except ValueError as e:
            logger.error(f"Not a number [{value}]")
            return False

        if (min_value and value_int < min_value) or (max_value and value_int > max_value):
            return False
        else:
            logger.info(f"Value [{value}] is between min_value [{min_value}] and [{max_value}]")
            return True

    @classmethod
    def setup_authentication(cls, profile="lab"):
        """
        Sets up the .vec_auth.yaml file in the users home directory in the "lab" profile.
        """
        logger = logging.getLogger(cls.__class__.__name__)

        loader = HomeConfigLoader(profile)
        if loader.has_profile():
            logger.info(f"We already have a configuration for [{profile}]")
            logger.info("If you proceed, THIS WILL BE OVERWRITTEN!!")
            valid_choice = False
            user_input = None
            while not valid_choice:
                user_input = str(input("Do you want to update existing profile? y/[n]").strip() or "n").lower()
                valid_choice = cls.valid_bool(user_input)
            if user_input == 'n':
                logger.info("Exiting")
                return
        else:
            logger.info("No existing profile found")

        def get_non_blank(field_name: str, sensitive=False) -> str:
            valid_field = False
            field_value = ""
            while not valid_field:
                message = f"Please enter the {field_name}:"
                if sensitive:
                    field_value = getpass.getpass(message)
                else:
                    field_value = str(input(message).strip())
                valid_field = len(field_value) > 0
                if not valid_field:
                    logger.error(f"Not a valid {field_name}: [{field_value}]")

            return field_value

        # Get a valid customerId.
        customer_id = get_non_blank("Customer Id")

        auth_question = "Please select authentication type:\n\t1. API Key (Default)\n\t 2. OAuth2\nEnter 1 or 2:"
        valid_choice = False
        user_input: str = ""
        while not valid_choice:
            user_input = input(auth_question).strip()
            valid_choice = cls.valid_int(user_input, min_value=1, max_value=2)
        choice = int(user_input)

        config: ClientConfig

        if choice == 1:
            api_key = get_non_blank("API Key", sensitive=True)
            api_key_auth_config = ApiKeyAuthConfig(api_key=api_key)
            config = ClientConfig(customer_id=customer_id, auth=api_key_auth_config)
        elif choice == 2:
            app_client_id = get_non_blank("Application Client Id")
            app_client_secret = get_non_blank("Application Client Secret", sensitive=True)
            oauth2_auth_config = OAuth2AuthConfig(app_client_id=app_client_id, app_client_secret=app_client_secret)
            config = ClientConfig(customer_id=customer_id, auth=oauth2_auth_config)
        else:
            raise Exception(f"Unexpected value: [{choice}]")

        logger.info(f"Saving configuration for profile [{profile}]")
        loader.save(config)

    def delete_labs(self, group_name:str, user_prefix=True):
        """
        Deletes all the labs with the given name match. We use the implicit username.
        :param group_name:
        :param user_prefix: Whether we should add the user prefix
        :return:
        """

        name, key = self._build_lab_name_and_key(group_name, user_prefix=user_prefix)
        self.logger.info(f"Deleting all lab corpora with matching name [{name}]")

        # Find all the corpora with group name, e.g. "getting-started" with user prefix.
        corpora = self.corpus_manager.find_corpora_with_filter(name)
        for corpus in corpora:
            self.logger.info(f"Deleting corpus [{corpus.name}] with key [{corpus.key}]")
            self.corpus_manager.delete(corpus.key)


class BaseFormatter(ABC):

    def __init__(self):
        pass

    def heading(self, heading: str, level: int = 1):
        raise Exception("Implement in subclass")

    def sentence(self, sentence: str):
        raise Exception("Implement in subclass")

    def link(self, text: str, url: str):
        raise Exception("Implement in subclass")

    def paragraph(self, paragraph: str):
        raise Exception("Implement in subclass")

    def bold(self, text: str):
        raise Exception("Implement in subclass")

    def italic(self, text: str):
        raise Exception("Implement in subclass")

    def list(self, items: List[str], level: int = 1):
        raise Exception("Implement in subclass")

    def rtl(self, text: str):
        raise Exception("Implement in subclass")

    def code(self, text: str, language=None):
        raise Exception("Implement in subclass")

class MarkdownFormatter(BaseFormatter):

    def __init__(self):
        super().__init__()

    def heading(self, heading: str, level: int = 1):
        indent = "#" * level
        return f'\n{indent} {heading}'

    def sentence(self, sentence: str):
        return sentence

    def link(self, text, url):
        return f"[{text}]({url})"

    def paragraph(self, paragraph: str):
        return f"\n\n{paragraph}\n"

    def bold(self, text: str):
        return f"**{text}**"

    def italic(self, text: str):
        return f"*{text}*"

    def list(self, items: List[str], level: int = 1):
        if level < 1:
            raise TypeError("List level must be greater than 0")
        indent = " " * (level - 1)
        results = [f"{indent} {idx + 1}. {item}" for idx, item in enumerate(items)]
        return "\n" + "\n".join(results) + "\n"

    def rtl(self, text):
        return '<div dir="rtl">\n' + text + '</div>'

    def code(self, text: str, language=None):
        result = "\n```"
        if language:
            result += language
        result += str(text)
        result += "\n```\n"
        return result

class CitationUtil:

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    def streamline(self, search_response: QueryFullResponse):
        """
        Streamlines a vectara response (starting with a specific response under the outer array)

        :param search_response: the first response inside the array response
        :return: a dict containing a "summary" with citations streamlined and a list of results used with ordering preserved
        """

        # TODO 1. Create a list of search results
        # TODO 2. Iterate through the summary, and record which search result is used at which index

        summary_text = search_response.summary

        if not summary_text:
            raise Exception("You should only call streamline for responses with a summary")
        find_result = re.finditer(r"\[(\d)\]", summary_text)

        index = 1
        dict_map: Dict[str, int] = {}

        # The end index of the last match.
        last_end = None

        results = []

        result_index = []

        for match_obj in find_result:
            match = match_obj.group(1)
            self.logger.debug(f"At {index} we found {match}")
            start = match_obj.start()
            end = match_obj.end()

            if match in dict_map:
                self.logger.debug(f"We already have result [{match}] at position [{dict_map[match]}]")
                pass
            else:
                self.logger.debug(f"Recording result [{match}] at position [{index}]")
                dict_map[match] = index
                result_index.append(int(match))
                index += 1

            if last_end:
                self.logger.debug("Using end of the last match to get last text chunk")
                results.append(summary_text[last_end:start])
            else:
                self.logger.debug("First iteration, use everything from beginning to start")
                if start > 0:
                    results.append(summary_text[0:start])

            results.append("[" + str(dict_map[match]) + "]")

            last_end = end

        if last_end and last_end < len(summary_text):
            results.append(summary_text[last_end:])

        result_summary = "".join(results)

        self.logger.debug(f"After transformation, we have the following summary:\n{result_summary}")

        self.logger.debug(f"The mapping to the result summary is as follows:\n{result_index}")

        streamline_response = {
            "summary": result_summary,
            "result_indexes": result_index
        }
        return streamline_response

class ResponseSetRenderer:

    def __init__(self, formatter: BaseFormatter):
        self.formatter = formatter
        self.logger = logging.getLogger(self.__class__.__name__)

    def render(self, query: str, response: QueryFullResponse, rtl=False, show_search_results=True, heading_level=1,
               expect_summary=True):
        f = self.formatter
        results = []

        # Build Heading
        results.append(f.heading(f"Query: {query}", level=heading_level))



        summary_text = None

        results_included = False

        # Build Summary
        if expect_summary:
            if response.summary:
                citation_util = CitationUtil()
                streamlined_response = citation_util.streamline(response)
                results.append(f.paragraph(streamlined_response["summary"]))

                if show_search_results:
                    docs = []
                    for id in streamlined_response["result_indexes"]:
                        if not response.search_results:
                            self.logger.warning(f"Search results are empty, can not reference result [{id}]")
                            continue
                        result = response.search_results[id]
                        doc_id = result.document_id
                        item = f.sentence(result.text) + " " + f.italic("score: " + str(result.score) + ", doc-id: " + doc_id)
                        docs.append(item)
                    list_text = f.list(docs)
                    results.append(list_text)
                    results_included = True
            else:
                summary_text = "We did not receive a summary, there are insufficient results returned from the corpus"
                results.append(f.paragraph(summary_text))

        # Build items
        if show_search_results and not results_included:
            docs = []
            if not response.search_results:
                self.logger.warning("Search results are empty, can not create final results")
            else:
                for index, result in enumerate(response.search_results):
                    doc_id = result.document_id

                    item = f.sentence(result.text) + " " + f.italic("score: " + str(result.score) + ", doc-id: " + doc_id)
                    docs.append(item)

            list_text = f.list(docs)
            results.append(list_text)

        results_text = "".join(results)

        if rtl:
            return f.rtl(results_text)
        else:
            return results_text


def render_markdown(query: str, response: QueryFullResponse, rtl=False, show_search_results=True, heading_level=1,
                    expect_summary=True):
    formatter = MarkdownFormatter()
    renderer = ResponseSetRenderer(formatter)
    return renderer.render(query, response, rtl, show_search_results, heading_level, expect_summary=expect_summary)

def render_markdown_req(request:dict):
    formatter = MarkdownFormatter()

    result = formatter.heading(f"Request Operation [{request['operation']}] with payload:", level=3)
    result += formatter.code(json.dumps(request, indent=4), "json")

    return result



# This file was auto-generated by Fern from our API Definition.

import typing
from ..core.client_wrapper import SyncClientWrapper
from ..types.corpus_key import CorpusKey
from .. import core
from ..types.components_schemas_max_chars_chunking_strategy import ComponentsSchemasMaxCharsChunkingStrategy
from ..types.table_extraction_config import TableExtractionConfig
from ..core.request_options import RequestOptions
from ..types.document import Document
from ..core.jsonable_encoder import jsonable_encoder
import json
from ..core.pydantic_utilities import parse_obj_as
from ..errors.bad_request_error import BadRequestError
from ..types.bad_request_error_body import BadRequestErrorBody
from ..errors.forbidden_error import ForbiddenError
from ..types.error import Error
from ..errors.not_found_error import NotFoundError
from ..types.not_found_error_body import NotFoundErrorBody
from json.decoder import JSONDecodeError
from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class UploadClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def file(
        self,
        corpus_key: CorpusKey,
        *,
        file: core.File,
        request_timeout: typing.Optional[int] = None,
        request_timeout_millis: typing.Optional[int] = None,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        chunking_strategy: typing.Optional[ComponentsSchemasMaxCharsChunkingStrategy] = OMIT,
        table_extraction_config: typing.Optional[TableExtractionConfig] = OMIT,
        filename: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Document:
        """
        Upload files such as PDFs and Word Documents for automatic text extraction and metadata parsing.
        The request expects a `multipart/form-data` format containing the following parts:
        * `metadata` - (Optional) Specifies a JSON object representing any additional metadata to be associated with the extracted document. For example, `'metadata={"key": "value"};type=application/json'`
        * `chunking_strategy` - (Optional) Specifies the chunking strategy for the platform to use. If you do not set this option, the platform uses the default strategy, which creates one chunk per sentence. For example, `'chunking_strategy={"type":"max_chars_chunking_strategy","max_chars_per_chunk":200};type=application/json'`
        * `table_extraction_config` - (Optional) Specifies whether to extract table data from the uploaded file. If you do not set this option, the platform does not extract tables from PDF files. Example config, `'table_extraction_config={"extract_tables":true};type=application/json'`
        * `file` - Specifies the file that you want to upload.
        * `filename` - Specified as part of the file field with the file name that you want to associate with the uploaded file. For a curl example, use the following syntax: `'file=@/path/to/file/file.pdf;filename=desired_filename.pdf'`

        For more detailed information, see this [File Upload API guide.](https://docs.vectara.com/docs/api-reference/indexing-apis/file-upload/file-upload)

        Parameters
        ----------
        corpus_key : CorpusKey
            The unique key identifying the corpus of which to upload the file.

        file : core.File
            See core.File for more documentation

        request_timeout : typing.Optional[int]
            The API will make a best effort to complete the request in the specified seconds or time out.

        request_timeout_millis : typing.Optional[int]
            The API will make a best effort to complete the request in the specified milliseconds or time out.

        metadata : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]
            Arbitrary object that will be attached as document metadata to the extracted document.

        chunking_strategy : typing.Optional[ComponentsSchemasMaxCharsChunkingStrategy]

        table_extraction_config : typing.Optional[TableExtractionConfig]

        filename : typing.Optional[str]
            Optional multipart section to override the filename.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Document
            The extracted document has been parsed and added to the corpus.

        Examples
        --------
        from vectara import Vectara

        client = Vectara(
            api_key="YOUR_API_KEY",
            client_id="YOUR_CLIENT_ID",
            client_secret="YOUR_CLIENT_SECRET",
        )
        client.upload.file(
            corpus_key="my-corpus",
        )
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v2/corpora/{jsonable_encoder(corpus_key)}/upload_file",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            data={},
            files={
                **(
                    {"metadata": (None, json.dumps(jsonable_encoder(metadata)), "application/json")}
                    if metadata is not OMIT
                    else {}
                ),
                **(
                    {"chunking_strategy": (None, json.dumps(jsonable_encoder(chunking_strategy)), "application/json")}
                    if chunking_strategy is not OMIT
                    else {}
                ),
                **(
                    {
                        "table_extraction_config": (
                            None,
                            json.dumps(jsonable_encoder(table_extraction_config)),
                            "application/json",
                        )
                    }
                    if table_extraction_config is not OMIT
                    else {}
                ),
                **(
                    {"filename": (None, json.dumps(jsonable_encoder(filename)), "text/plain")}
                    if filename is not OMIT
                    else {}
                ),
                "file": core.with_content_type(file=file, default_content_type="application/octet-stream"),
            },
            headers={
                "Request-Timeout": str(request_timeout) if request_timeout is not None else None,
                "Request-Timeout-Millis": str(request_timeout_millis) if request_timeout_millis is not None else None,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    Document,
                    parse_obj_as(
                        type_=Document,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 400:
                raise BadRequestError(
                    typing.cast(
                        BadRequestErrorBody,
                        parse_obj_as(
                            type_=BadRequestErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncUploadClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def file(
        self,
        corpus_key: CorpusKey,
        *,
        file: core.File,
        request_timeout: typing.Optional[int] = None,
        request_timeout_millis: typing.Optional[int] = None,
        metadata: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        chunking_strategy: typing.Optional[ComponentsSchemasMaxCharsChunkingStrategy] = OMIT,
        table_extraction_config: typing.Optional[TableExtractionConfig] = OMIT,
        filename: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> Document:
        """
        Upload files such as PDFs and Word Documents for automatic text extraction and metadata parsing.
        The request expects a `multipart/form-data` format containing the following parts:
        * `metadata` - (Optional) Specifies a JSON object representing any additional metadata to be associated with the extracted document. For example, `'metadata={"key": "value"};type=application/json'`
        * `chunking_strategy` - (Optional) Specifies the chunking strategy for the platform to use. If you do not set this option, the platform uses the default strategy, which creates one chunk per sentence. For example, `'chunking_strategy={"type":"max_chars_chunking_strategy","max_chars_per_chunk":200};type=application/json'`
        * `table_extraction_config` - (Optional) Specifies whether to extract table data from the uploaded file. If you do not set this option, the platform does not extract tables from PDF files. Example config, `'table_extraction_config={"extract_tables":true};type=application/json'`
        * `file` - Specifies the file that you want to upload.
        * `filename` - Specified as part of the file field with the file name that you want to associate with the uploaded file. For a curl example, use the following syntax: `'file=@/path/to/file/file.pdf;filename=desired_filename.pdf'`

        For more detailed information, see this [File Upload API guide.](https://docs.vectara.com/docs/api-reference/indexing-apis/file-upload/file-upload)

        Parameters
        ----------
        corpus_key : CorpusKey
            The unique key identifying the corpus of which to upload the file.

        file : core.File
            See core.File for more documentation

        request_timeout : typing.Optional[int]
            The API will make a best effort to complete the request in the specified seconds or time out.

        request_timeout_millis : typing.Optional[int]
            The API will make a best effort to complete the request in the specified milliseconds or time out.

        metadata : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]
            Arbitrary object that will be attached as document metadata to the extracted document.

        chunking_strategy : typing.Optional[ComponentsSchemasMaxCharsChunkingStrategy]

        table_extraction_config : typing.Optional[TableExtractionConfig]

        filename : typing.Optional[str]
            Optional multipart section to override the filename.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        Document
            The extracted document has been parsed and added to the corpus.

        Examples
        --------
        import asyncio

        from vectara import AsyncVectara

        client = AsyncVectara(
            api_key="YOUR_API_KEY",
            client_id="YOUR_CLIENT_ID",
            client_secret="YOUR_CLIENT_SECRET",
        )


        async def main() -> None:
            await client.upload.file(
                corpus_key="my-corpus",
            )


        asyncio.run(main())
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v2/corpora/{jsonable_encoder(corpus_key)}/upload_file",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            data={},
            files={
                **(
                    {"metadata": (None, json.dumps(jsonable_encoder(metadata)), "application/json")}
                    if metadata is not OMIT
                    else {}
                ),
                **(
                    {"chunking_strategy": (None, json.dumps(jsonable_encoder(chunking_strategy)), "application/json")}
                    if chunking_strategy is not OMIT
                    else {}
                ),
                **(
                    {
                        "table_extraction_config": (
                            None,
                            json.dumps(jsonable_encoder(table_extraction_config)),
                            "application/json",
                        )
                    }
                    if table_extraction_config is not OMIT
                    else {}
                ),
                **(
                    {"filename": (None, json.dumps(jsonable_encoder(filename)), "text/plain")}
                    if filename is not OMIT
                    else {}
                ),
                "file": core.with_content_type(file=file, default_content_type="application/octet-stream"),
            },
            headers={
                "Request-Timeout": str(request_timeout) if request_timeout is not None else None,
                "Request-Timeout-Millis": str(request_timeout_millis) if request_timeout_millis is not None else None,
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return typing.cast(
                    Document,
                    parse_obj_as(
                        type_=Document,  # type: ignore
                        object_=_response.json(),
                    ),
                )
            if _response.status_code == 400:
                raise BadRequestError(
                    typing.cast(
                        BadRequestErrorBody,
                        parse_obj_as(
                            type_=BadRequestErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    typing.cast(
                        Error,
                        parse_obj_as(
                            type_=Error,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    typing.cast(
                        NotFoundErrorBody,
                        parse_obj_as(
                            type_=NotFoundErrorBody,  # type: ignore
                            object_=_response.json(),
                        ),
                    )
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class RerankedSearchResult(UniversalBaseModel):
    """
    An individual reranked search result from a query.
    """

    text: typing.Optional[str] = pydantic.Field(default=None)
    """
    The document part altered by the context configuration that matches the query.
    """

    score: typing.Optional[float] = pydantic.Field(default=None)
    """
    The score of the individual result.
    """

    original_score: typing.Optional[float] = pydantic.Field(default=None)
    """
    The original score of the individual result before reranking.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .individual_search_result import IndividualSearchResult
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class StreamSearchResponse(UniversalBaseModel):
    """
    The search response results.
    """

    type: typing.Literal["search_results"] = "search_results"
    search_results: typing.Optional[typing.List[IndividualSearchResult]] = pydantic.Field(default=None)
    """
    The ranked search results.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import pydantic
import typing
from .api_operation_policy import ApiOperationPolicy
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class ApiPolicy(UniversalBaseModel):
    """
    What actions a principal can take on the Vectara platform.
    """

    name: str = pydantic.Field()
    """
    The name of the API role.
    """

    allowed_operations: typing.Dict[str, ApiOperationPolicy] = pydantic.Field()
    """
    Operations that are allowed by the API role. Each operation may only allow
    certain resources that are described by a resource path to resource value
    map. If the value is null, then the operation is allowed on any resource.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from vectara.documents.client import DocumentsClient
from vectara.types import Document, CoreDocument, StructuredDocument
from vectara.utils.hash import calculate_sha256
from typing import Optional, Dict, Any, Union, Tuple
from enum import Enum
import logging

HASH_FIELD = "sha256"

class DocOpEnum(Enum):
    CREATED = 1
    UPDATED = 2
    IGNORED = 3

class DocumentManager:

    def __init__(self, documents_client: DocumentsClient):
        self.documents_client = documents_client
        self.logger = logging.getLogger(self.__class__.__name__)

    def check_exists(self, corpus_key: str, doc_id: str) -> Optional[Document]:
        """
        Checks for a document by Id, returning that Document if it exists, otherwise returning None.

        :param corpus_key: The corpus we expect the document to exist in.
        :param doc_id: The ID of the document
        :return: The found document or None
        """
        def list_documents_gen():
            response = self.documents_client.list(corpus_key, metadata_filter=f"doc.id = '{doc_id}'")

            for item in response:
                yield item

        for document in list_documents_gen():
            return document

        return None

    def check_same(self, corpus_key: str, doc_id: str, content: bytes, metadata: Optional[Dict[str, Any]] = None) -> Tuple[bool, Optional[bool]]:
        """
        Checks whether the corpus contains the existing document with a matching SHA256 hash and same metadata attributes.
        
        This will only check the metadata attributes listed in the method signature, if there are additional metadata
        attributes on the document in the corpus, these will be ignored.
        
        :param corpus_key: the corpus which holds the document
        :param doc_id: the id of the document we are checking
        :param content: the byte contents of the document we are checking.
        :param metadata: metadata fields to validate.
        :return: A boolean indicator that the document was found, has the same content and listed metadata attributes match
        """

        doc = self.check_exists(corpus_key, doc_id)
        if not doc:
            return False, None

        if doc.metadata and (HASH_FIELD not in doc.metadata or type(doc.metadata[HASH_FIELD]) is not str):
            return True, False

        # TODO Validate that document has matching metadata
        if metadata:
            if len(metadata.keys()) > 0 and doc.metadata is None:
                return True, False

            for key in metadata.keys():
                value = metadata[key]
                if value is None:
                    raise TypeError("Cannot compare a metadata attribute of value None")

                if not doc.metadata or key not in doc.metadata:
                    # Either we don't have any existing metadata or the key isn't present.
                    return True, False

                existing = doc.metadata[key]

                if existing != value:
                    return True, False

        if not doc.metadata or HASH_FIELD not in doc.metadata:
            # Existing document does not have hash field set.
            return True, False

        existing_hash: str = str(doc.metadata[HASH_FIELD])
        current_hash: str = calculate_sha256(content)
        if existing_hash.lower() != current_hash.lower():
            # Existing document has different value for hash field.
            return True, False
        else:
            return True, True


    def index_doc(self, corpus_key: str, doc: Union[CoreDocument, StructuredDocument]) -> DocOpEnum:

        # Remove the sha256 hash from metadata.
        if doc.metadata and HASH_FIELD in doc.metadata:
            del doc.metadata[HASH_FIELD]

        content = doc.model_dump_json().encode("utf-8")

        # Check exists and whether same.
        exists, same = self.check_same(corpus_key, doc.id, content, doc.metadata)

        def set_metadata():
            sha256_hash = calculate_sha256(content)
            if doc.metadata:
                metadata_copy = dict(doc.metadata)
                metadata_copy[HASH_FIELD] = sha256_hash
                result = doc.copy(update={"metadata": metadata_copy})
            else:
                result = doc.copy(update={"metadata": {HASH_FIELD: sha256_hash}})

            return result

        if exists and same:
            self.logger.info("Document already exists with same hash, skipping")
            return DocOpEnum.IGNORED

        doc = set_metadata()

        if not exists:
            self.logger.info("Document doesnt exist, creating fresh")
            self.documents_client.create(corpus_key, request=doc)
            return DocOpEnum.CREATED

        if exists and not same:
            self.logger.info(f"Document with id [{doc.id}] exists deleting existing and creating fresh (upsert)")
            self.documents_client.delete(corpus_key, doc.id)
            self.documents_client.create(corpus_key, request=doc)
            return DocOpEnum.UPDATED
        else:
            raise Exception("Invalid combination of exists/same, should not get here")












# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class ModelParameters(UniversalBaseModel):
    """
    The parameters for the model.
    WARNING: This is an experimental feature, and breakable at any point with virtually no
    notice. It is meant for experimentation to converge on optimal parameters that can then
    be set in the prompt definitions.
    """

    max_tokens: typing.Optional[int] = pydantic.Field(default=None)
    """
    The maximum number of tokens to be returned by the model.
    """

    temperature: typing.Optional[float] = pydantic.Field(default=None)
    """
    The sampling temperature to use. Higher values make the output more random, while lower
    values make it more focused and deterministic.
    """

    frequency_penalty: typing.Optional[float] = pydantic.Field(default=None)
    """
    Higher values penalize new tokens based on their existing frequency in the text so far,
    decreasing the model's likelihood to repeat the same line verbatim.
    """

    presence_penalty: typing.Optional[float] = pydantic.Field(default=None)
    """
    Higher values penalize new tokens based on whether they appear in the text so far,
    increasing the model's likelihood to talk about new topics.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class DocumentStorageUsage(UniversalBaseModel):
    """
    How much storage the document used. This information is currently not returned when
    retrieving the document, and only returned when indexing a document.
    """

    bytes_used: typing.Optional[int] = pydantic.Field(default=None)
    """
    Number of bytes used by document counting towards maximum corpus size, and
    towards any billing plans.
    """

    metadata_bytes_used: typing.Optional[int] = pydantic.Field(default=None)
    """
    Number of metadata bytes used by a document.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class GenerationInfo(UniversalBaseModel):
    """
    Event containing information on how the generation was accomplished.
    """

    type: typing.Literal["generation_info"] = "generation_info"
    rendered_prompt: typing.Optional[str] = pydantic.Field(default=None)
    """
    The rendered prompt sent to the LLM. Useful when creating customer `prompt_template` templates.
    """

    rephrased_query: typing.Optional[str] = pydantic.Field(default=None)
    """
    View the actual query made to backend that was rephrased 
    by the LLM from the input query.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
